package com.jclark.xsl;

import java.io.IOException;
import java.net.URL;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

class SheetImpl implements Sheet {
  private
  interface TopLevelParser {
    void parse(Node node) throws XSLException, IOException;
  }

  private
  class IncludeParser implements TopLevelParser {
    public void parse(Node ruleNode) throws XSLException, IOException {
      parseTopLevel(getSheetNode(parser.load(new URL(ruleNode.getURL(),
			                             getRequiredAttribute(ruleNode,
						                          HREF)))));
    }
  }

  private
  class ImportParser implements TopLevelParser {
    public void parse(Node ruleNode) throws XSLException, IOException {
      if (imports == null)
	imports = new Vector();
      imports.addElement(new SheetImpl(new URL(ruleNode.getURL(),
					       getRequiredAttribute(ruleNode, HREF)),
				       parser));
    }
  }

  private
  class RuleParser implements TopLevelParser {
    public void parse(Node ruleNode) throws XSLException {
      try {
	templateRules.add(PatternParser
			  .createMatchers(ruleNode,
					  getOptionalAttribute(ruleNode,
							       MATCH,
							       "/")),
			  Integer.parseInt(getOptionalAttribute(ruleNode,
								PRIORITY,
								"0")),
			  parseActions(ruleNode, emptyAction));
      }
      catch (NumberFormatException e) {
	throw new XSLException("invalid priority", ruleNode);
      }
    }
  }

  private
  class AttributeRuleParser implements TopLevelParser {
    public void parse(Node ruleNode) throws XSLException {
      try {
	attributeRules.add(PatternParser.createMatchers(ruleNode,
							getRequiredAttribute(ruleNode, MATCH)),
			   Integer.parseInt(getOptionalAttribute(ruleNode,
								 PRIORITY,
								 "0")),
			   getAttributeSet(ruleNode));
      }
      catch (NumberFormatException e) {
	throw new XSLException("invalid priority", ruleNode);
      }
    }
  }

  private
  class DefineAttributeSetParser implements TopLevelParser {
    public void parse(Node defNode) throws XSLException {
      // FIXME deal with duplicates
      attributeSetTable.put(defNode.getNamespacePrefixMap().expandElementTypeName(getRequiredAttribute(defNode, NAME), defNode),
			    getAttributeSet(defNode));
    }
  }

  private
  class DefineMacroParser implements TopLevelParser {
    public void parse(Node defNode) throws XSLException {
      MacroArgumentList args = new MacroArgumentList();
      // FIXME deal with duplicates
      macroTable.put(defNode.getNamespacePrefixMap().expandElementTypeName(getRequiredAttribute(defNode, NAME), defNode),
		     new MacroImpl(args,
				   parseArgsThenActions(defNode, emptyAction, args, true)));
    }
  }

  private
  class DefineConstantParser implements TopLevelParser {
    public void parse(Node defNode) throws XSLException {
      // FIXME deal with duplicates
      constantExpressionTable
	.put(defNode.getNamespacePrefixMap().expandElementTypeName(getRequiredAttribute(defNode, NAME), defNode),
	     PatternParser.createAttributeValueTemplate(defNode,
			                                getRequiredAttribute(defNode, VALUE)));
    }
  }

  private
  class IdParser implements TopLevelParser {
    public void parse(Node node) throws XSLException {
      // FIXME deal with element attribute
      idAttributeList.addElement(getRequiredAttribute(node, ATTRIBUTE));
    }
  }

  private
  class StripSpaceParser implements TopLevelParser {
    public void parse(Node node) throws XSLException {
      // FIXME Implement
    }
  }

  private
  class PreserveSpaceParser extends StripSpaceParser {
      // FIXME Implement
  }


  private
  interface ActionParser {
    Action parse(Node node) throws XSLException;
  }

  private
  class ProcessChildrenParser implements ActionParser {
    public Action parse(Node node) {
      return processChildrenAction;
    }
  }

  private
  class ContentsParser implements ActionParser {
    public Action parse(Node node) {
      return new ContentsAction();
    }
  }

  private
  class ProcessParser implements ActionParser {
    public Action parse(Node node) throws XSLException {
      return new ProcessAction(PatternParser.createQuery(node,
							 getRequiredAttribute(node, SELECT)));
    }
  }

  private
  class ForEachParser implements ActionParser {
    public Action parse(Node node) throws XSLException {
      return new ForEachAction(PatternParser
			       .createQuery(node, getRequiredAttribute(node, SELECT)),
			       parseActions(node, emptyAction));
    }
  }

  private
  class IfParser implements ActionParser {
    public Action parse(Node node) throws XSLException {
      return new IfAction(makeCondition(node),
			  parseActions(node, emptyAction),
			  emptyAction);
    }

    Condition makeCondition(Node node) throws XSLException {
      return new QueryCondition(PatternParser.createQuery(node,
	                                                  getRequiredAttribute(node, TEST)));
    }
  }

  private
  class InvokeParser implements ActionParser {
    public Action parse(Node node) throws XSLException {

      MacroArgumentList args = new MacroArgumentList();
      return new InvokeAction(node.getNamespacePrefixMap().expandElementTypeName(getRequiredAttribute(node, MACRO), node),
			      args,
			      parseArgsThenActions(node, emptyAction, args, false));
    }
  }

  private
  class ChooseParser extends IfParser {
    public Action parse(Node node) throws XSLException {
      return parseChoices(node.getChildren(true));
    }
    public Action parseChoices(NodeIterator iter) throws XSLException {
      Node node = iter.next();
      if (node == null)
	return emptyAction;
      Name name = node.getName();
      if (name.equals(XSL_WHEN))
	return new IfAction(makeCondition(node),
			    parseActions(node, emptyAction),
			    parseChoices(iter));
      if (name.equals(XSL_OTHERWISE)) {
	Node next = iter.next();
	if (next != null)
	  throw new XSLException("unexpected element after otherwise",
				 next);
	return parseActions(node, emptyAction);
      }
      throw new XSLException("expected when or otherwise", node);
    }
  }

  private
  class TextParser implements ActionParser {
    public Action parse(Node node) throws XSLException {
      Node child = node.getChildren(false).next();
      if (child != null) {
	String data = child.getData();
	if (data == null)
	  throw new XSLException("xsl:text must not contain elements");
	return new CharsAction(data);
      }
      else
	return emptyAction;
    }
  }

  private
  class ValueOfParser implements ActionParser {
    public Action parse(Node node) throws XSLException {
      return new ValueOfAction(PatternParser.createExpression(node, getRequiredAttribute(node, EXPR)));
    }
  }

  private
  class NumberParser implements ActionParser {
    public Action parse(Node node) throws XSLException {
      String level = getOptionalAttribute(node, LEVEL, "single");
      String countString = node.getAttributeValue(COUNT);
      Matchable count;
      if (countString == null)
	count = null;
      else
	count = PatternParser.createMatchable(node, countString);
      String fromString = node.getAttributeValue(FROM);
      Matchable from;
      if (fromString == null)
	from = null;
      else
	from = PatternParser.createMatchable(node, fromString);
      NumberListFormatTemplate format = getNumberListFormatTemplate(node);
      if (level.equals("any"))
	return new AnyLevelNumberAction(count, from, format);
      if (level.equals("multi"))
	return new MultiLevelNumberAction(count, from, format);
      return new SingleLevelNumberAction(count, from, format);
    }
  }

  class ProcessContextImpl implements ProcessContext {
    Node root;
    Hashtable constantValueTable = new Hashtable();
    ProcessContextImpl(Node root) {
      this.root = root;
    }
    public
    void process(Node node, Result result) throws XSLException {
      // FIXME Detect processing loops.
      Action action = getAction(this, node);
      if (action == null)
	action = processChildrenAction;
      AttributeListTemplate attributesTemplate = applyAttributeRules(this, node);
      if (attributesTemplate != null)
	result.element(sequenceElementTypeName,
	               attributesTemplate.instantiate(this, node),
		       null,
		       action,
		       this,
		       node);
      else
	action.invoke(this, node, result);
    }

    public
    boolean getStripSource(Node parentNode) {
      // FIXME for strip-space and preserve-space
      return stripSource;
    }

    public
    Macro getMacro(Name name) {
      return SheetImpl.this.getMacro(name);
    }

    public
    String getConstant(Name name) throws XSLException {
      String value = (String)constantValueTable.get(name);
      if (value != null)
	return value;
      // Avoid possibility of infinite loop
      constantValueTable.put(name, "");
      Expression expr = getConstantExpression(name);
      if (expr == null)
	value = "";
      else
	value = expr.eval(this, root);
      constantValueTable.put(name, value);
      return value;
    }

    public
    AttributeListTemplate getNamedAttributeSet(Name name) {
      return SheetImpl.this.getNamedAttributeSet(name);
    }

    public
    void processContents(Result result) throws XSLException {
      // FIXME should catch this
    }

    public
    String getMacroArgument(Name name) {
      return null;
    }

    public
    boolean hasId(Node node, String name) {
      return hasAttribute(idAttributeList, node, name);
    }

    public
    Node getElementWithId(String name) {
      // FIXME build table of IDs
      return null;
    }

    public final
    boolean hasAttribute(Vector nameList, Node node, String value) {
      int len = nameList.size();
      for (int i = 0; i < len; i++)
	if (value.equals(node.getAttributeValue((Name)nameList.elementAt(i))))
	  return true;
      return false;
    }

  }


  boolean stripSource = false;
  private boolean indentResult = false;
  private String resultNamespace = null;
  private TemplateRuleSet templateRules = new TemplateRuleSet();
  private AttributeRuleSet attributeRules = new AttributeRuleSet();
  private Hashtable macroTable = new Hashtable();
  private Hashtable constantExpressionTable = new Hashtable();
  // Named attribute sets
  private Hashtable attributeSetTable = new Hashtable();
  private Vector idAttributeList = new Vector();
  private static final Action emptyAction = new EmptyAction();
  private static final Action processChildrenAction = new ProcessAction(new ChildrenQuery());
  private Hashtable keywordTable = new Hashtable();
  private Vector imports = null;
  XMLProcessor parser;
  Name XSL_WHEN;
  Name XSL_OTHERWISE;
  Name XSL_STYLESHEET;
  Name XSL_ARG;
  Name XSL_MACRO_ARG;
  Name XSL_ATTRIBUTE_SET;
  Name HREF;
  Name MATCH;
  Name PRIORITY;
  Name SELECT;
  Name TEST;
  Name DEFAULT_SPACE;
  Name INDENT_RESULT;
  Name RESULT_NS;
  Name NAME;
  Name DEFAULT;
  Name VALUE;
  Name MACRO;
  Name ATTRIBUTE;
  Name EXPR;
  Name FROM;
  Name COUNT;
  Name LEVEL;
  Name FORMAT;
  Name LETTER_VALUE;
  Name XML_LANG;
  Name N_DIGITS_PER_GROUP;
  Name DIGIT_GROUP_SEP;
  Name sequenceElementTypeName;
  
  SheetImpl(URL url, XMLProcessor parser) throws IOException, XSLException {
    this(parser);
    parseSheet(parser.load(url));
  }

  private
  SheetImpl(XMLProcessor parser) {
    this.parser = parser;
    XSL_WHEN = xsl("when");
    XSL_OTHERWISE = xsl("otherwise");
    XSL_STYLESHEET = xsl("stylesheet");
    XSL_ARG = xsl("arg");
    XSL_MACRO_ARG = xsl("macro-arg");
    XSL_ATTRIBUTE_SET = xsl("attribute-set");
    HREF = parser.createName("href");
    SELECT = parser.createName("select");
    TEST = parser.createName("test");
    MATCH = parser.createName("match");
    PRIORITY = parser.createName("priority");
    DEFAULT_SPACE = parser.createName("default-space");
    INDENT_RESULT = parser.createName("indent-result");
    RESULT_NS = parser.createName("result-ns");
    NAME = parser.createName("name");
    DEFAULT = parser.createName("default");
    VALUE = parser.createName("value");
    MACRO = parser.createName("macro");
    ATTRIBUTE = parser.createName("attribute");
    EXPR = parser.createName("expr");
    COUNT = parser.createName("count");
    LEVEL = parser.createName("level");
    FROM = parser.createName("from");
    FORMAT = parser.createName("format");
    LETTER_VALUE = parser.createName("letter-value");
    XML_LANG = parser.createName("xml:lang", Name.XML_NAMESPACE);
    N_DIGITS_PER_GROUP = parser.createName("n-digits-per-group");
    DIGIT_GROUP_SEP = parser.createName("digit-group-sep");
    sequenceElementTypeName = parser.createName("fo:sequence", XSL_NAMESPACE + "/FO");
    keywordTable.put(xsl("include"), new IncludeParser());
    keywordTable.put(xsl("import"), new ImportParser());
    keywordTable.put(xsl("template"), new RuleParser());
    keywordTable.put(xsl("apply"), new AttributeRuleParser());
    keywordTable.put(xsl("define-attribute-set"), new DefineAttributeSetParser());
    keywordTable.put(xsl("define-macro"), new DefineMacroParser());
    keywordTable.put(xsl("define-constant"), new DefineConstantParser());
    keywordTable.put(xsl("id"), new IdParser());
    keywordTable.put(xsl("text"), new TextParser());
    keywordTable.put(xsl("process"), new ProcessParser());
    keywordTable.put(xsl("process-children"), new ProcessChildrenParser());
    keywordTable.put(xsl("contents"), new ContentsParser());
    keywordTable.put(xsl("for-each"), new ForEachParser());
    keywordTable.put(xsl("value-of"), new ValueOfParser());
    keywordTable.put(xsl("number"), new NumberParser());
    keywordTable.put(xsl("if"), new IfParser());
    keywordTable.put(xsl("choose"), new ChooseParser());
    keywordTable.put(xsl("invoke"), new InvokeParser());
    keywordTable.put(xsl("strip-space"), new StripSpaceParser());
    keywordTable.put(xsl("preserve-space"), new PreserveSpaceParser());
    templateRules.add(new TextMatcher(), 0, new DataAction());
  }

  public
  Result process(Node node, Result root) throws XSLException {
    new ProcessContextImpl(node).process(node, root);
    root.finish();
    return root;
  }

  public
  boolean getIndentResult() {
    return indentResult;
  }

  public
  String getResultNamespace() {
    return resultNamespace;
  }

  AttributeListTemplate applyAttributeRules(ProcessContext context, Node node) throws XSLException {
    AttributeListTemplate attribute = attributeRules.getAttributes(context, node);
    if (imports != null) {
      ComposeAttributeListTemplate compoundAttribute = null;
      for (int i = imports.size(); --i >= 0;) {
	AttributeListTemplate tem = ((SheetImpl)imports.elementAt(i)).applyAttributeRules(context, node);
	if (tem != null) {
	  if (compoundAttribute != null)
	    compoundAttribute.add(tem);
	  else if (attribute != null)
	    attribute = compoundAttribute = new ComposeAttributeListTemplate(attribute, tem);
	  else
	    attribute = tem;
	}
      }
    }
    return attribute;
  }

  Action getAction(ProcessContext context, Node node) throws XSLException {
    Action action = templateRules.getAction(context, node);
    if (imports != null) {
      for (int i = imports.size(); action == null && --i >= 0;)
	action = ((SheetImpl)imports.elementAt(i)).getAction(context, node);
    }
    return action;
  }

  Macro getMacro(Name name) {
    Macro macro = (Macro)macroTable.get(name);
    if (macro != null)
      return macro;
    if (imports != null) {
      for (int i = imports.size(); --i >= 0;) {
	macro = ((SheetImpl)imports.elementAt(i)).getMacro(name);
	if (macro != null)
	  return macro;
      }
    }
    return null;
  }

  Expression getConstantExpression(Name name) {
    Expression expr = (Expression)constantExpressionTable.get(name);
    if (expr != null)
      return expr;
    if (imports != null) {
      for (int i = imports.size(); --i >= 0;) {
	expr = ((SheetImpl)imports.elementAt(i)).getConstantExpression(name);
	if (expr != null)
	  return expr;
      }
    }
    return null;
  }

  AttributeListTemplate getNamedAttributeSet(Name name) {
    AttributeListTemplate attributeList
      = (AttributeListTemplate)attributeSetTable.get(name);
    if (attributeList != null)
      return attributeList;
    if (imports != null) {
      for (int i = imports.size(); --i >= 0;) {
	attributeList = ((SheetImpl)imports.elementAt(i)).getNamedAttributeSet(name);
	if (attributeList != null)
	  return attributeList;
      }
    }
    return null;
  }

  Node getSheetNode(Node rootNode) throws XSLException {
    Node sheetNode = rootNode.getChildren(true).next();
    if (!XSL_STYLESHEET.equals(sheetNode.getName())) {
      if (sheetNode.getName().toString().equals("xsl:stylesheet"))
  	throw new XSLException("declare 'xsl:' namespace with <xsl:stylesheet xsl:='http://www.w3.org/TR/WD-xsl'>", sheetNode);
      else
       throw new XSLException("bad document element", sheetNode);
    }
    return sheetNode;
  }

  private
  void parseSheet(Node rootNode) throws XSLException, IOException {
    Node sheetNode = getSheetNode(rootNode);
    if ("strip".equals(sheetNode.getAttributeValue(DEFAULT_SPACE)))
      stripSource = true;
    if ("yes".equals(sheetNode.getAttributeValue(INDENT_RESULT)))
      indentResult = true;
    resultNamespace = sheetNode.getAttributeValue(RESULT_NS);
    if (resultNamespace != null) {
      NamespacePrefixMap map = sheetNode.getNamespacePrefixMap();
      if (resultNamespace.equals(""))
	resultNamespace = map.getDefaultNamespace();
      else
	resultNamespace = map.getNamespace(resultNamespace);
    }
    parseTopLevel(sheetNode);
    templateRules.sort();
  }

  void parseTopLevel(Node sheetNode) throws XSLException, IOException {
    for (NodeIterator iter = sheetNode.getChildren(true);;) {
      Node node = iter.next();
      if (node == null)
	break;
      TopLevelParser parser = null;
      try {
	parser = (TopLevelParser)keywordTable.get(node.getName());
      }
      catch (ClassCastException e) { }
      if (parser == null)
	throw new XSLException("bad top-level element", node);
      parser.parse(node);
    }
  }

  Action parseActions(Node node, Action ifEmpty) throws XSLException {
    return parseArgsThenActions(node, ifEmpty, null, false);
  }
  
  Action parseArgsThenActions(Node node, Action ifEmpty, MacroArgumentList args, boolean defining) throws XSLException {
    NodeIterator iter = node.getChildren(true);
    node = iter.next();
    for (;; node = iter.next()) {
      if (node == null)
	return ifEmpty;
      if (args == null)
	break;
      Name name = node.getName();
      if (!name.equals(defining ? XSL_MACRO_ARG : XSL_ARG))
	break;
      args.add(node.getNamespacePrefixMap().expandAttributeName(getRequiredAttribute(node, NAME), node),
	       defining
	       ? getOptionalAttribute(node, DEFAULT, "")
	       : getRequiredAttribute(node, VALUE),
	       node);
    }
    AppendAction sequence = new AppendAction();
    do {
      switch (node.getType()) {
      case Node.TEXT:
	sequence.add(new CharsAction(node.getData()));
	break;
      case Node.ELEMENT:
	{
	  Name name = node.getName();
	  if (!XSL_NAMESPACE.equals(name.getNamespace()))
	    sequence.add(new LiteralElementAction(unquoteName(node.getName(), node),
					          new AttributeListTemplateImpl(node),
					          unquoteNamespacePrefixMap(node.getNamespacePrefixMap()),
					          parseActions(node, null)));
	  else {
	    ActionParser actionParser = null;
	    try {
	      actionParser = (ActionParser)keywordTable.get(name);
	    }
	    catch (ClassCastException e) { }
	    if (actionParser == null)
	      throw new XSLException("expected action not " + name, node);
	    sequence.add(actionParser.parse(node));
	  }
	}
      }
      node = iter.next();
    } while (node != null);
    return sequence;
  }

  String getRequiredAttribute(Node node, Name name) throws XSLException {
    String value = node.getAttributeValue(name);
    if (value == null)
      throw new XSLException("missing attribute " + name, node);
    return value;
  }

  String getOptionalAttribute(Node node, Name name, String dflt) {
    String value = node.getAttributeValue(name);
    return value == null ? dflt : value;
  }

  String getData(Node node) throws XSLException {
    node = node.getChildren(true).next();
    if (node == null)
      return "";
    String data = node.getData();
    if (data == null)
      throw new XSLException("only character data allowed", node);
    return data;
  }

  AttributeListTemplate getAttributeSet(Node node) throws XSLException {
    NodeIterator iter = node.getChildren(true);
    Node attributeSetNode = iter.next();
    if (attributeSetNode == null
	|| !attributeSetNode.getName().equals(XSL_ATTRIBUTE_SET)
	|| iter.next() != null)
      throw new XSLException("content must be a single xsl:attribute-set element", node);
    return new AttributeListTemplateImpl(attributeSetNode);
  }

  NumberListFormatTemplate getNumberListFormatTemplate(Node node) throws XSLException {
    NumberListFormatTemplate t = new NumberListFormatTemplate();
    String value = node.getAttributeValue(FORMAT);
    if (value != null) {
      Expression expr = PatternParser.createAttributeValueTemplate(node, value);
      if (expr instanceof ConstantExpression)
	t.setFormat(((ConstantExpression)expr).getConstantValue());
      else
	t.setFormat(expr);
    }
    // FIXME should be able to use attribute value templates for these
    value = node.getAttributeValue(XML_LANG);
    if (value != null)
      t.setLang(value);
    value = node.getAttributeValue(LETTER_VALUE);
    if (value != null)
      t.setLetterValue(value);
    value = node.getAttributeValue(N_DIGITS_PER_GROUP);
    if (value != null) {
      try {
	t.setNDigitsPerGroup(Integer.parseInt(value));
      }
      catch (NumberFormatException e) { }
    }
    value = node.getAttributeValue(DIGIT_GROUP_SEP);
    if (value != null)
      t.setDigitGroupSep(value);
    // FIXME handle sequenceSrc
    return t;
  }

  static final
  String XSL_NAMESPACE = "http://www.w3.org/TR/WD-xsl";

  private
  Name xsl(String name) {
    return parser.createName("xsl:" + name,  XSL_NAMESPACE);
  }

  static
  Name unquoteName(Name name, Node node) {
    String ns = unquoteNamespace(name.getNamespace());
    if (ns == null || ns.length() == 0)
      return name;
    return node.getNamespacePrefixMap().createName(name.toString(), ns);
  }

  static private
  NamespacePrefixMap unquoteNamespacePrefixMap(NamespacePrefixMap map) {
    String ns = unquoteNamespace(map.getDefaultNamespace());
    if (ns != null) {
      if (ns.length() == 0)
	map = map.unbindDefault();
      else
	map = map.bindDefault(ns);
    }
    int size = map.getSize();
    NamespacePrefixMap newMap = map;
    for (int i = 0; i < size; i++) {
      ns = unquoteNamespace(map.getNamespace(i));
      if (ns != null) {
        if (ns.length() == 0)
	  newMap = newMap.unbind(map.getPrefix(i));
	else
	  newMap = newMap.bind(map.getPrefix(i), ns);
      }
    }
    return newMap;
  }

  static private
  String unquoteNamespace(String ns) {
    if (ns == null)
      return null;
    if (!ns.startsWith(XSL_NAMESPACE))
      return null;
    if (ns.equals(XSL_NAMESPACE))
      return "";
    for (int i = XSL_NAMESPACE.length();;) {
      if (!ns.regionMatches(i, "/quote", 0, 6))
	break;
      i += 6;
      if (i == ns.length())
	return ns.substring(0, ns.length() - 6);
    }
    return null;
  }

}
